use crate::errors::UtilesCoreResult;
use crate::UtilesCoreError;
use std::num::ParseIntError;
use std::ops::BitAnd;

use crate::UtilesCoreError::InvalidZoom;

type Zooms = Vec<u8>;

/// ZoomSet is a set of zoom levels represented as a 32 bit unsigned integer
/// where each bit represents a zoom level (0 to 31). The least significant bit
/// represents zoom level 31 and the most significant bit represents zoom level 0.
///
/// # Examples
/// ```
/// use utiles_core::zoom::ZoomSet;
/// let zset = ZoomSet::new(0b11100000_00000000_00000000_00000000);
/// let zooms_vec: Vec<u8> = zset.into();
/// assert_eq!(zooms_vec, vec![0, 1, 2]);
/// ```
#[derive(Debug, Clone, Eq, PartialEq, Default)]
pub struct ZoomSet(u32);

/// Return a vector of zoom levels from a zoom-set u32
#[must_use]
pub fn zset2zvec(zset: u32) -> Vec<u8> {
    (0..32)
        .rev()
        .filter(|&i| (zset & (1 << i)) != 0)
        .map(|i| (31 - i) as u8) // Convert bit index to zoom level and cast to u8
        .collect()
}

/// Return a zoom-set u32 from a vector of zoom levels
#[must_use]
pub fn zvec2zset(zvec: Vec<u8>) -> u32 {
    zvec.iter().fold(0, |acc, &z| acc | (1 << (31 - z)))
}

/// ZoomSet is a set of zoom levels represented as a u32
///
/// # Examples
/// ```
/// use utiles_core::zoom::ZoomSet;
/// let zset = ZoomSet::new(0b11100000_00000000_00000000_00000000);
/// let zooms_vec: Vec<u8> = zset.into();
/// assert_eq!(zooms_vec, vec![0, 1, 2]);
/// ```
impl ZoomSet {
    /// Create a new ZoomSet from a u32
    #[must_use]
    pub fn new(zset: u32) -> Self {
        Self(zset)
    }

    /// Create a new ZoomSet from a vector of zoom levels
    #[must_use]
    pub fn from_zooms(zooms: Vec<u8>) -> Self {
        Self(zvec2zset(zooms))
    }

    /// Return a vector of zoom levels from a zoom-set u32
    #[must_use]
    pub fn to_zooms(&self) -> Vec<u8> {
        zset2zvec(self.0)
    }
}

impl From<u8> for ZoomSet {
    fn from(zoom: u8) -> Self {
        ZoomSet(1 << (31 - zoom))
    }
}

impl From<u32> for ZoomSet {
    fn from(zset: u32) -> Self {
        ZoomSet(zset)
    }
}

impl TryFrom<Vec<u8>> for ZoomSet {
    type Error = UtilesCoreError;

    fn try_from(zvec: Vec<u8>) -> Result<Self, Self::Error> {
        let result = zvec.iter().try_fold(0u32, |acc, &z| {
            if z > 31 {
                Err(InvalidZoom(z.to_string()))
            } else {
                Ok(acc | (1 << (31 - z)))
            }
        })?;
        Ok(ZoomSet::new(result)) // Replace with actual construction method
    }
}

impl BitAnd for ZoomSet {
    type Output = ZoomSet;

    fn bitand(self, rhs: Self) -> Self::Output {
        ZoomSet(self.0 & rhs.0)
    }
}

impl BitAnd<u32> for ZoomSet {
    type Output = ZoomSet;

    fn bitand(self, rhs: u32) -> Self::Output {
        ZoomSet(self.0 & rhs)
    }
}

impl BitAnd<ZoomSet> for u32 {
    type Output = ZoomSet;

    fn bitand(self, rhs: ZoomSet) -> Self::Output {
        ZoomSet(self & rhs.0)
    }
}

impl BitAnd<u8> for ZoomSet {
    type Output = ZoomSet;

    fn bitand(self, rhs: u8) -> Self::Output {
        ZoomSet(self.0 & (1 << (31 - rhs)))
    }
}

impl From<ZoomSet> for Vec<u8> {
    fn from(zset: ZoomSet) -> Self {
        zset2zvec(zset.0)
    }
}

type ZoomsSetInt = u32;

/// Struct representing zoom level range (min, max)
#[derive(Debug, Clone, Eq, PartialEq)]
pub struct ZoomRange {
    /// Minimum zoom level
    pub min: u8,
    /// Maximum zoom level
    pub max: u8,
}

// default zoom range
impl Default for ZoomRange {
    fn default() -> Self {
        Self { min: 0, max: 31 }
    }
}

impl ZoomRange {
    /// Create a new ZoomRange
    #[must_use]
    pub fn new(min: u8, max: u8) -> Self {
        Self { min, max }
    }

    /// Create a new ZoomRange from a maximum zoom level (0 to max)
    #[must_use]
    pub fn from_max(max: u8) -> Self {
        Self { min: 0, max }
    }

    /// Create a new ZoomRange from a minimum zoom level (min to 31)
    #[must_use]
    pub fn from_min(min: u8) -> Self {
        Self { min, max: 31 }
    }
}

/// convert range of zoom levels to a set of zoom levels
///
/// # Examples
/// ```
/// use utiles_core::zoom::{ZoomRange, ZoomSet};
/// let zrange = ZoomRange::new(0, 7);
/// let zset: ZoomSet = zrange.into();
/// assert_eq!(zset, ZoomSet::new(0b11111111_00000000_00000000_00000000));
/// ```
impl From<ZoomRange> for ZoomSet {
    fn from(zoom_range: ZoomRange) -> Self {
        ZoomSet(
            (zoom_range.min..=zoom_range.max)
                .rev()
                .fold(0, |acc, z| acc | (1 << (31 - z))),
        )
    }
}

/// parse a string of zoom levels to a vector of zoom levels
/// # Examples
/// ```
/// use utiles_core::zoom::parse_zooms;
/// let zvec = parse_zooms("0,1,2,3,4,5,6,7");
/// assert_eq!(zvec.unwrap(), vec![0, 1, 2, 3, 4, 5, 6, 7]);
/// ```
///
/// ```
/// use utiles_core::zoom::parse_zooms;
/// let zvec = parse_zooms("0-7");
/// assert_eq!(zvec.unwrap(), vec![0, 1, 2, 3, 4, 5, 6, 7]);
/// ```
///
/// ```
/// use utiles_core::zoom::parse_zooms;
/// let zvec = parse_zooms("0-2,4-7");
/// assert_eq!(zvec.unwrap(), vec![0, 1, 2, 4, 5, 6, 7]);
/// ```
///
/// ```
/// use utiles_core::zoom::parse_zooms;
/// let zvec = parse_zooms("1,2,3,4,4,4,4,4");
/// assert_eq!(zvec.unwrap(), vec![1, 2, 3, 4]);
/// ```
pub fn parse_zooms(zstr: &str) -> UtilesCoreResult<Vec<u8>> {
    let mut zvec: Vec<u8> = vec![];
    for z in zstr.split(',') {
        if z.contains('-') {
            let zrange: Result<Vec<u8>, ParseIntError> = z
                .split('-')
                .map(|z| z.parse::<u8>())
                .collect::<Result<Vec<_>, _>>();

            let zrange = match zrange {
                Ok(zrange) => match zrange.len() {
                    1 => vec![zrange[0]],
                    2 => (zrange[0]..=zrange[1]).collect(),
                    _ => vec![],
                },
                Err(_) => return Err(InvalidZoom(z.to_string())),
            };
            zvec.extend(zrange);
        } else {
            match z.parse::<u8>() {
                Ok(num) => zvec.push(num),
                Err(_) => return Err(InvalidZoom(z.to_string())),
            }
        }
    }
    // make sure zooms are between 0 and 32
    for z in &zvec {
        if *z > 32 {
            return Err(InvalidZoom((*z).to_string()));
        }
    }
    // unique and sort zooms
    zvec.sort();
    zvec.dedup();
    Ok(zvec)
}

/// Enum representing a single zoom level or a vec of zoom levels
pub enum ZoomOrZooms {
    /// A single zoom level
    Zoom(u8),

    /// A vec of zoom levels
    Zooms(Vec<u8>),
}

impl From<u8> for ZoomOrZooms {
    fn from(zoom: u8) -> Self {
        ZoomOrZooms::Zoom(zoom)
    }
}

impl From<Vec<u8>> for ZoomOrZooms {
    fn from(zooms: Vec<u8>) -> Self {
        ZoomOrZooms::Zooms(zooms)
    }
}

impl From<ZoomOrZooms> for Zooms {
    fn from(zoom_or_zooms: ZoomOrZooms) -> Self {
        match zoom_or_zooms {
            ZoomOrZooms::Zoom(zoom) => vec![zoom],
            ZoomOrZooms::Zooms(zooms) => zooms,
        }
    }
}

impl From<ZoomOrZooms> for ZoomsSetInt {
    fn from(zoom_or_zooms: ZoomOrZooms) -> Self {
        match zoom_or_zooms {
            ZoomOrZooms::Zoom(zoom) => 1 << (31 - zoom),
            ZoomOrZooms::Zooms(zooms) => zvec2zset(zooms),
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn zset2zvec_none() {
        let zset: u32 = 0b00000000_00000000_00000000_00000000; // Example, zoom levels 2 and 4 are set
        let zvec: Vec<u8> = vec![];
        assert_eq!(zset2zvec(zset), zvec);
    }

    #[test]
    fn zset2zvec_0_1_2() {
        let zset: u32 = 0b11100000_00000000_00000000_00000000; // Example, zoom levels 2 and 4 are set
        let zvec: Vec<u8> = vec![0, 1, 2];

        assert_eq!(zset2zvec(zset), zvec);
    }

    #[test]
    fn zset2zvec_all() {
        let zset: u32 = 0b11111111_11111111_11111111_11111111; // Example, zoom levels 2 and 4 are set
        let zvec: Vec<u8> = vec![
            0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
            21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
        ];
        assert_eq!(zset2zvec(zset), zvec);
    }

    #[test]
    fn zvec2zset_none() {
        let zset: u32 = 0b00000000_00000000_00000000_00000000; // Example, zoom levels 2 and 4 are set
        let zvec: Vec<u8> = vec![];
        assert_eq!(zvec2zset(zvec), zset);
    }

    #[test]
    fn zvec2zset_0_1_2() {
        let zset: u32 = 0b11100000_00000000_00000000_00000000; // Example, zoom levels 2 and 4 are set
        let zvec: Vec<u8> = vec![0, 1, 2];

        assert_eq!(zvec2zset(zvec), zset);
    }

    #[test]
    fn zvec2zset_all() {
        let zset: u32 = 0b11111111_11111111_11111111_11111111; // Example, zoom levels 2 and 4 are set
        let zvec: Vec<u8> = vec![
            0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
            21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
        ];
        assert_eq!(zvec2zset(zvec), zset);
    }

    #[test]
    fn zvec2zset_0_1_2_3_4_5_6_7() {
        let zset: u32 = 0b11111111_00000000_00000000_00000000; // Example, zoom levels 2 and 4 are set
        let zvec: Vec<u8> = vec![0, 1, 2, 3, 4, 5, 6, 7];
        let zset_from_zvec = zvec2zset(zvec);
        assert_eq!(zset_from_zvec, zset);
    }

    #[test]
    fn zoom_set_into_zoom_vec() {
        let zset_int: u32 = 0b11111111_00000000_00000000_00000000; // Example, zoom levels 2 and 4 are set
        let zset: ZoomSet = zset_int.into();
        let zvec = Vec::from(zset);
        assert_eq!(zvec, vec![0, 1, 2, 3, 4, 5, 6, 7]);

        // let zvec: Vec<u8> = vec![0, 1, 2, 3, 4, 5, 6, 7];
        // let zset_from_zvec = zvec2zset(zvec);
        // assert_eq!(zset_from_zvec, zset);
    }
}
